@@ -8,10 +8,11 @@ import codecs
 import json
 from collections import OrderedDict
 from datetime import datetime
-from typing import List, Optional
+from typing import Any, Dict, List, Optional
 
 from .fields import (
     ByteField,
+    ByteOrder,
     BytesField,
     DateTimeField,
     Field,
@@ -19,9 +20,7 @@ from .fields import (
     FloatingPointField,
     NumericArrayField,
     NumericField,
-    hi_to_lo,
-    lo_to_hi,
-    lo_to_hi_signed,
+    Sign,
 )
 
 
@@ -39,30 +38,28 @@ class Packet(object):
         serial_number: int,
         seconds: int,
         pulse_counts: List[int],
-        temperatures: Optional[int],
+        temperatures: List[float],
         polarized_watt_seconds: Optional[List[int]] = None,
-        currents: Optional[float] = None,
+        currents: Optional[List[float]] = None,
         time_stamp: Optional[datetime] = None,
-        **kwargs
+        **kwargs: Dict[str, Any]
     ):
-        self.packet_format = packet_format
-        self.voltage = voltage
-        self.absolute_watt_seconds = absolute_watt_seconds
-        if polarized_watt_seconds:
-            self.polarized_watt_seconds = polarized_watt_seconds
-        if currents:
-            self.currents = currents
-        self.device_id = device_id
-        self.serial_number = serial_number
-        self.seconds = seconds
-        self.pulse_counts = pulse_counts
-        self.temperatures = temperatures
+        self.packet_format: PacketFormat = packet_format
+        self.voltage: float = voltage
+        self.absolute_watt_seconds: List[int] = absolute_watt_seconds
+        self.polarized_watt_seconds: Optional[List[int]] = polarized_watt_seconds
+        self.currents: Optional[List[float]] = currents
+        self.device_id: int = device_id
+        self.serial_number: int = serial_number
+        self.seconds: int = seconds
+        self.pulse_counts: List[int] = pulse_counts
+        self.temperatures: List[float] = temperatures
         if time_stamp:
-            self.time_stamp = time_stamp
+            self.time_stamp: datetime = time_stamp
         else:
-            self.time_stamp = datetime.now()
+            self.time_stamp: datetime = datetime.now()
 
-    def __str__(self):
+    def __str__(self) -> str:
         return json.dumps(
             {
                 "device_id": self.device_id,
@@ -70,10 +67,8 @@ class Packet(object):
                 "seconds": self.seconds,
                 "voltage": self.voltage,
                 "absolute_watt_seconds": self.absolute_watt_seconds,
-                "polarized_watt_seconds": self.polarized_watt_seconds
-                if hasattr(self, "polarized_watt_seconds")
-                else None,
-                "currents": self.currents if hasattr(self, "currents") else None,
+                "polarized_watt_seconds": self.polarized_watt_seconds,
+                "currents": self.currents,
                 "pulse_counts": self.pulse_counts,
                 "temperatures": self.temperatures,
                 "time_stamp": self.time_stamp.isoformat(),
@@ -81,36 +76,47 @@ class Packet(object):
         )
 
     @property
-    def type(self):
+    def type(self) -> str:
         return self.packet_format.name
 
     @property
-    def num_channels(self):
+    def num_channels(self) -> int:
         return self.packet_format.num_channels
 
-    @property
-    def max_seconds(self):
-        assert isinstance(self.packet_format.fields["seconds"], NumericField)
-        return self.packet_format.fields["seconds"].max
+    def delta_seconds(self, prev: int) -> int:
+        field = self.packet_format.fields["seconds"]
+        assert isinstance(field, NumericField)
+        return self._delta_value(field, self.seconds, prev)
 
-    @property
-    def max_pulse_count(self):
-        assert isinstance(self.packet_format.fields["pulse_counts"], NumericArrayField)
-        return self.packet_format.fields["pulse_counts"].elem_field.max
+    def delta_pulse_count(self, index: int, prev: int) -> int:
+        field = self.packet_format.fields["pulse_counts"]
+        assert isinstance(field, NumericArrayField)
+        return self._delta_value(field.elem_field, self.pulse_counts[index], prev)
 
-    @property
-    def max_absolute_watt_seconds(self):
-        assert isinstance(
-            self.packet_format.fields["absolute_watt_seconds"], NumericArrayField
+    def delta_absolute_watt_seconds(self, index: int, prev: int) -> int:
+        field = self.packet_format.fields["absolute_watt_seconds"]
+        assert isinstance(field, NumericArrayField)
+        return self._delta_value(
+            field.elem_field, self.absolute_watt_seconds[index], prev
         )
-        return self.packet_format.fields["absolute_watt_seconds"].elem_field.max
 
-    @property
-    def max_polarized_watt_seconds(self):
-        assert isinstance(
-            self.packet_format.fields["polarized_watt_seconds"], NumericArrayField
-        )
-        return self.packet_format.fields["polarized_watt_seconds"].elem_field.max
+    def delta_polarized_watt_seconds(self, index: int, prev: int) -> int:
+        field = self.packet_format.fields["polarized_watt_seconds"]
+        assert isinstance(field, NumericArrayField)
+        if self.polarized_watt_seconds is not None:
+            return self._delta_value(
+                field.elem_field, self.polarized_watt_seconds[index], prev
+            )
+        else:
+            return 0
+
+    def _delta_value(self, field: NumericField, cur: int, prev: int) -> int:
+        if prev > cur:
+            diff = field.max + 1 - prev
+            diff += cur
+        else:
+            diff = cur - prev
+        return diff
 
 
 class PacketFormat(object):
@@ -122,46 +128,49 @@ class PacketFormat(object):
         name: str,
         num_channels: int,
         has_net_metering: bool = False,
-        has_time_stamp=False,
+        has_time_stamp: bool = False,
     ):
-        self.name = name
-        self.num_channels = num_channels
+        self.name: str = name
+        self.num_channels: int = num_channels
         self.fields: OrderedDict[str, Field] = OrderedDict()
 
-        self.fields["header"] = NumericField(3, hi_to_lo)
-        self.fields["voltage"] = FloatingPointField(2, hi_to_lo, 10.0)
+        self.fields["header"] = NumericField(3, ByteOrder.HiToLo, Sign.Unsigned)
+        self.fields["voltage"] = FloatingPointField(
+            2, ByteOrder.HiToLo, Sign.Unsigned, 10.0
+        )
         self.fields["absolute_watt_seconds"] = NumericArrayField(
-            num_channels, 5, lo_to_hi
+            num_channels, 5, ByteOrder.LoToHi, Sign.Unsigned
         )
         if has_net_metering:
             self.fields["polarized_watt_seconds"] = NumericArrayField(
-                num_channels, 5, lo_to_hi
+                num_channels, 5, ByteOrder.LoToHi, Sign.Unsigned
             )
-        self.fields["serial_number"] = NumericField(2, hi_to_lo)
+        self.fields["serial_number"] = NumericField(2, ByteOrder.HiToLo, Sign.Unsigned)
         self.fields["reserved"] = ByteField()
-        self.fields["device_id"] = NumericField(1, hi_to_lo)
+        self.fields["device_id"] = NumericField(1, ByteOrder.HiToLo, Sign.Unsigned)
         self.fields["currents"] = FloatingPointArrayField(
-            num_channels, 2, lo_to_hi, 50.0
+            num_channels, 2, ByteOrder.LoToHi, Sign.Unsigned, 50.0
         )
-        self.fields["seconds"] = NumericField(3, lo_to_hi)
+        self.fields["seconds"] = NumericField(3, ByteOrder.LoToHi, Sign.Unsigned)
         self.fields["pulse_counts"] = NumericArrayField(
-            PacketFormat.NUM_PULSE_COUNTERS, 3, lo_to_hi
+            PacketFormat.NUM_PULSE_COUNTERS, 3, ByteOrder.LoToHi, Sign.Unsigned
         )
         self.fields["temperatures"] = FloatingPointArrayField(
             PacketFormat.NUM_TEMPERATURE_SENSORS,
             2,
-            lo_to_hi_signed,
+            ByteOrder.LoToHi,
+            Sign.Signed,
             2.0,
         )
         if num_channels == 32:
             self.fields["spare_bytes"] = BytesField(2)
         if has_time_stamp:
             self.fields["time_stamp"] = DateTimeField()
-        self.fields["footer"] = NumericField(2, hi_to_lo)
+        self.fields["footer"] = NumericField(2, ByteOrder.HiToLo, Sign.Unsigned)
         self.fields["checksum"] = ByteField()
 
     @property
-    def size(self):
+    def size(self) -> int:
         result = 0
         for value in self.fields.values():
             result += value.size
